/*
 * SuffixAutomaton.h
 *
 *  Created on: Jan 8, 2017
 *      Author: Jan Travnicek
 */

#ifndef SUFFIX_AUTOMATON_H_
#define SUFFIX_AUTOMATON_H_

#include <alib/string>
#include <alib/set>
#include <alib/iostream>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>
#include <primitive/Unsigned.h>
#include <primitive/xml/Unsigned.h>

#include <alphabet/common/SymbolNormalize.h>

#include <automaton/FSM/DFA.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class SuffixAutomaton final : public object::ObjectBase {
protected:
	automaton::DFA < SymbolType, unsigned > m_automaton;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit SuffixAutomaton ( automaton::DFA < SymbolType, unsigned > automaton );

	/**
	 * @return Root node of the trie
	 */
	const automaton::DFA < SymbolType, unsigned > & getAutomaton ( ) const &;

	automaton::DFA < SymbolType, unsigned > && getAutomaton ( ) &&;

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return m_automaton.getInputAlphabet ( );
	}

	ext::set < SymbolType > && getAlphabet ( ) && {
		return std::move ( m_automaton ).getInputAlphabet ( );
	}

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return m_automaton.removeInputSymbol ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const SuffixAutomaton & other ) const;

	virtual explicit operator std::string ( ) const;

	virtual object::ObjectBase * inc ( ) &&;

	typedef SuffixAutomaton < > normalized_type;
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
SuffixAutomaton < SymbolType >::SuffixAutomaton ( automaton::DFA < SymbolType, unsigned > automaton ) : m_automaton ( std::move ( automaton ) ) {
}

template < class SymbolType >
object::ObjectBase * SuffixAutomaton < SymbolType >::clone ( ) const {
	return new SuffixAutomaton ( * this );
}

template < class SymbolType >
object::ObjectBase * SuffixAutomaton < SymbolType >::plunder ( ) && {
	return new SuffixAutomaton ( std::move ( * this ) );
}

template < class SymbolType >
const automaton::DFA < SymbolType, unsigned > & SuffixAutomaton < SymbolType >::getAutomaton ( ) const & {
	return m_automaton;
}

template < class SymbolType >
automaton::DFA < SymbolType, unsigned > && SuffixAutomaton < SymbolType >::getAutomaton ( ) && {
	return std::move ( m_automaton );
}

template < class SymbolType >
void SuffixAutomaton < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(SuffixAutomaton " << this->m_automaton << ")";
}

template < class SymbolType >
int SuffixAutomaton < SymbolType >::compare ( const SuffixAutomaton & other ) const {
	auto first = ext::tie ( getAutomaton ( ) );
	auto second = ext::tie ( other.getAutomaton ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
SuffixAutomaton < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
object::ObjectBase* SuffixAutomaton < SymbolType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace core {

template < class SymbolType >
struct normalize < indexes::stringology::SuffixAutomaton < SymbolType >, typename std::enable_if < ! std::is_same < indexes::stringology::SuffixAutomaton < SymbolType >, indexes::stringology::SuffixAutomaton < > >::value >::type > {
	static indexes::stringology::SuffixAutomaton < > eval ( indexes::stringology::SuffixAutomaton < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlphabet ( ) );
		ext::vector < DefaultSymbolType > string = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( value ).getString ( ) );

		return indexes::stringology::SuffixAutomaton < > ( std::move ( alphabet ), std::move ( value ).getData ( ), std::move ( string ) );
	}
};

template < class SymbolType >
struct xmlApi < indexes::stringology::SuffixAutomaton < SymbolType > > {
	static indexes::stringology::SuffixAutomaton < SymbolType > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::stringology::SuffixAutomaton < SymbolType > & data );
};

template < class SymbolType >
indexes::stringology::SuffixAutomaton < SymbolType > xmlApi < indexes::stringology::SuffixAutomaton < SymbolType > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	automaton::DFA < SymbolType, unsigned > automaton = core::xmlApi < automaton::DFA < SymbolType, unsigned > >::parse ( input );
	indexes::stringology::SuffixAutomaton < SymbolType > res ( std::move ( automaton ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType >
bool xmlApi < indexes::stringology::SuffixAutomaton < SymbolType > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType >
const std::string & xmlApi < indexes::stringology::SuffixAutomaton < SymbolType > >::xmlTagName ( ) {
	static std::string xmlTagName = "SuffixAutomaton";

	return xmlTagName;
}

template < class SymbolType >
void xmlApi < indexes::stringology::SuffixAutomaton < SymbolType > >::compose ( ext::deque < sax::Token > & output, const indexes::stringology::SuffixAutomaton < SymbolType > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	core::xmlApi < automaton::DFA < SymbolType, unsigned > >::compose ( output, index.getAutomaton ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace core */

#endif /* SUFFIX_AUTOMATON_H_ */
